"""Paster"""

from os.path import dirname, exists
import promus.util as util


def paste(prs, name):
    "Paste a template to current location. "
    tmp = dirname(__file__)
    if exists('%s/%s' % (tmp, name)):
        util.exec_cmd('cp -r %s/%s/* ./' % (tmp, name), True)
    else:
        prs.dismiss("PASTE>> no template found...", 1)
    prs.dismiss("PASTE>> done...", 0)



"""Send

Sends a request to add collaborators or to invite collaborators to
a particular repository.

"""

import os
import socket
import textwrap
import promus.core.ssh as ssh
import promus.core.git as git
import promus.core.util as util

DESC = """
To be able to collaborate first you need to let your collaborators be
able to connect to your account. You may do this by sending a request.

    promus send request email@hostname 'First Last'

On the other hand, if you already have a collaborator and you have
added his/her email or username to one of the acls in a repository;
then you may send them an invite so that they know how to connect to
the repository.

    promus send invite email@hostname

or

    promus send invite username

Please note that to send an invite you must call the command from the
repository, i.e. your current working directory must be a repository.

"""


def add_parser(subp, raw):
    "Add a parser to the main subparser. "
    tmpp = subp.add_parser('send', help='send a request or invite',
                           formatter_class=raw,
                           description=textwrap.dedent(DESC))
    tmpp.add_argument('type', metavar='TYPE', type=str,
                      choices=['request', 'invite'],
                      help='One of the following: request, invite')
    tmpp.add_argument('email', type=str,
                      help="collaborators email")
    tmpp.add_argument('name', type=str, nargs='?',
                      help="collaborators name")

REQUEST_TXT = """Hello {name},

This email has been generated on behalf of {master} to request for
your public key. This will give access to future repositories hosted
by {master} so that you may work as a team.

To accomplish this, you will need to have promus installed. If you do
not have it please visit the following page and take a moment to
learn the basics of promus.

    http://promus.readthedocs.org

Once you have promus you need to download the attached file. Then
navigate to the directory where you downloaded the file and type:

    promus add host {masteruser}@{host}

You and {master} will recieve a notification shortly after running
the command verifying that the connection was successful.

- Promus

"""
REQUEST_HTML = """<p>Hello {name},</p>
<p>This email has been generated on behalf of {master} to request for
your public key. This will give access to future repositories hosted
by {master} so that you may work as a team.</p>
<p>To accomplish this, you will need to have promus installed. If you
do not have it please visit the following page and take a moment to
learn the basics of promus.</p>
<pre><code>    <a
href="http://promus.readthedocs.org">http://promus.readthedocs.org</a>
</code></pre>
<p>Once you have promus you need to download the attached file. Then
navigate to the directory where you downloaded the file and type:</p>
<pre><code>    promus add host {masteruser}@{host}
</code></pre>
<p>You and {master} will recieve a notification shortly after
running the command verifying that the connection was successful.</p>
<p>
<strong>- Promus</strong>
</p>
"""


def send_request(arg):
    """Sends an email to request for a public key. """
    home = os.environ['HOME']
    host = socket.gethostname()
    master = os.environ['USER']
    key = ssh.make_key('%s/.promus/%s@%s' % (home, master, host))
    users, pending, unknown = ssh.read_authorized_keys()
    key_type, ssh_key = ssh.get_public_key(key).split()
    pending[ssh_key] = [arg.email, key_type, arg.email]
    ssh.write_authorized_keys(users, pending, unknown)
    if arg.name is None:
        name = 'future collaborator'
    else:
        name = arg.name
    mastername = git.config('user.name')
    util.send_mail([arg.email],
                   'Collaboration request from %s' % mastername,
                   REQUEST_TXT.format(name=name, masteruser=master,
                                      master=mastername, host=host),
                   REQUEST_HTML.format(name=name, masteruser=master,
                                       master=mastername, host=host),
                   [key])
    os.remove(key)


def send_invite(arg):
    """Sends instructions on how to connect to a repository. """
    pass


def run(arg):
    """Run command. """
    func = {
        'request': send_request,
        'invite': send_invite,
    }
    func[arg.type](arg)



"""SSH Utilities"""

import os
import re
import sys
import shutil
from os.path import exists
from promus.command import exec_cmd, date, error
PC = sys.modules['promus.core']

RE_USER = re.compile('command="python -m promus greet '
                     '\'(?P<email>.*?),(?P<user>.*?),'
                     '(?P<name>.*?),(?P<alias>.*?)\'" '
                     '(?P<type>.*?) (?P<key>.*?) (?P<desc>.*)')
RE_PENDING = re.compile('command="python -m promus add user '
                        '(?P<email>.*?)" (?P<type>.*?) (?P<key>.*?) '
                        '(?P<desc>.*)')


def make_key(name, cmt=''):
    """Creates a new key. """
    if not exists(name):
        cmd = "ssh-keygen -f %s -C '%s' -N '' -t rsa -q" % (name, cmt)
        exec_cmd(cmd, True)
    return name


def get_keys():
    """Verifies that the keys exist, if not it creates them. Returns
    the path of the keys. """
    alias = PC.config('host.alias')
    if alias == '':
        raise NameError("run `promus setup` to provide an alias")
    home = os.environ['HOME']
    master = os.environ['USER']
    # DEFAULT KEY
    idfname = '%s/.ssh/id_dsa' % home
    if not exists(idfname):
        idfname = '%s/.ssh/id_rsa' % home
        if not exists(idfname):
            cmt = '%s-%s' % (master, alias)
            cmd = "ssh-keygen -f %s -C '%s' -N '' -t rsa -q" % (idfname, cmt)
            exec_cmd(cmd, True)
    # GIT KEY
    fname = '%s/.ssh/%s-%s-git' % (home, master, alias)
    if not exists(fname):
        cmt = '%s-%s-git' % (master, alias)
        cmd = "ssh-keygen -f %s -C '%s' -N '' -t rsa -q" % (fname, cmt)
        exec_cmd(cmd, True)
    return idfname, fname


def get_public_key(private):
    """Retrieve the public key belonging to the given private key. """
    public, _, _ = exec_cmd('ssh-keygen -y -f %s' % private)
    return public.strip()


def read_config():
    """Read the ssh configuration file and return a dictionary with
    its information. """
    config = dict()
    try:
        configf = open('%s/.ssh/config' % os.environ['HOME'], 'r')
    except IOError:
        return config
    for line in configf:
        line = line.split()
        if not line or line[0] == '#':
            continue
        try:
            if line[0] == 'Host':
                entry = ' '.join(line[1:])
                config[entry] = dict()
            else:
                config[entry][line[0]] = line[1]
        except IndexError:
            error("ERROR: check ~/.ssh/config for errors\n")
    configf.close()
    return config


def write_config(config):
    """Write ssh configuration file. """
    last = list()
    configf = open('%s/.ssh/config' % os.environ['HOME'], 'w')
    configf.write('# config file generated on %s\n' % date())
    for entry in config:
        if 'IdentityFile' in config[entry]:
            last.append(entry)
        else:
            configf.write('Host %s\n' % entry)
            for key in config[entry]:
                configf.write('    %s %s\n' % (key, config[entry][key]))
            configf.write('\n')
    for entry in last:
        configf.write('Host %s\n' % entry)
        for key in config[entry]:
            configf.write('    %s %s\n' % (key, config[entry][key]))
        configf.write('\n')
    configf.close()
    exec_cmd('chmod 700 %s/.ssh/config' % os.environ['HOME'], True)


def read_authorized_keys():
    """Read the authorized keys file. """
    users = dict()
    pending = dict()
    unknown = list()
    try:
        usersf = open('%s/.ssh/authorized_keys' % os.environ['HOME'], 'r')
    except IOError:
        return users, pending, unknown
    for line in usersf:
        match = RE_USER.match(line)
        if match:
            if match.group('email') not in users:
                users[match.group('email')] = dict()
            content = [match.group('user'),
                       match.group('name'),
                       match.group('alias'),
                       match.group('type'),
                       match.group('desc')]
            users[match.group('email')][match.group('key')] = content
        else:
            match = RE_PENDING.match(line)
            if match:
                content = [match.group('email'),
                           match.group('type'),
                           match.group('desc')]
                pending[match.group('key')] = content
            else:
                tmp = line.strip()
                if tmp != '' and tmp[0] != '#':
                    unknown.append(line)
    usersf.close()
    return users, pending, unknown


def write_authorized_keys(users, pending, unknown):
    """Rewrite the authorized keys file. """
    ak_file = '%s/.ssh/authorized_keys' % os.environ['HOME']
    backup = '%s/.ssh/authorized_keys.promus-backup' % os.environ['HOME']
    if not os.path.exists(backup):
        shutil.copy(ak_file, backup)
    akfp = open(ak_file, 'w')
    akfp.write('# PROMUS: authorized_keys generated on %s\n' % date())
    emails = sorted(users.keys())
    for user in emails:
        for key, content in users[user].items():
            akfp.write('command="python -m promus greet \'%s,' % user)
            akfp.write('%s,%s,%s\'" ' % (content[0], content[1], content[2]))
            akfp.write('%s %s %s\n' % (content[3], key, content[4]))
    keys = pending.keys()
    if keys:
        akfp.write('# pending requests:\n')
        data = [(key, pending[key][0], pending[key][1]) for key in keys]
        data = sorted(data, key=lambda x: x[1])
        for item in data:
            akfp.write('command="python -m promus add user %s"' % item[1])
            akfp.write(' %s %s %s\n' % (item[2], item[0], item[1]))
    if unknown:
        akfp.write('# unknown keys:\n')
        for entry in unknown:
            akfp.write('%s' % entry)
    akfp.close()
    exec_cmd('chmod 700 %s' % ak_file, True)



""" Command line use of promus

To run promus from the command line do the following:

    python -m promus [-h] subcommand

Use the option --help for more information.

"""

import argparse
import textwrap
from os.path import split, abspath
from glob import iglob
from promus.__version__ import VERSION
from promus.command import import_mod
try:
    import argcomplete
except ImportError:
    pass


def parse_options(mod):
    """Interpret the command line inputs and options. """
    desc = """
promus is a remote manager designed to create and manage `git`
repositories in a remote server without the need of administrator
privileges.

use the help option with a command for more information on what
promus can do.

"""
    ver = "promus %s" % VERSION
    epi = """
more info:
  http://promus.readthedocs.org

version:
  promus %s

""" % VERSION
    raw = argparse.RawDescriptionHelpFormatter
    argp = argparse.ArgumentParser(formatter_class=raw,
                                   description=textwrap.dedent(desc),
                                   epilog=textwrap.dedent(epi))
    argp.add_argument('-v', '--version', action='version', version=ver)
    subp = argp.add_subparsers(title='subcommands',
                               dest='parser_name',
                               help='additional help',
                               metavar="<command>")
    names = sorted(mod.keys())
    for name in names:
        mod[name].add_parser(subp, raw)
    try:
        argcomplete.autocomplete(argp)
    except NameError:
        pass
    return argp.parse_args()


def run():
    """Run promus from the command line. """
    mod = dict()
    rootpath = split(abspath(__file__))[0]

    mod_names = [name for name in iglob('%s/command/*.py' % rootpath)]
    for name in mod_names:
        tmp_name = split(name)[1][:-3]
        tmp_mod = import_mod('promus.command.%s' % tmp_name)
        if hasattr(tmp_mod, 'add_parser'):
            mod[tmp_name] = tmp_mod

    arg = parse_options(mod)
    mod[arg.parser_name].run(arg)

if __name__ == '__main__':
    run()



"""Show

Display information about the keys and users that have access to your
account.

"""

import os
import re
import sys
import socket
import textwrap
import promus.core.ssh as ssh
import promus.core.git as git

RE_LINE = re.compile('(?P<stuff>.*?)ssh-(?P<type>.*?) '
                     '(?P<key>.*?) (?P<desc>.*)')
DESC = """
use this command to quickly view your ssh keys or the users that are
allowed to connect to your account.

"""


def add_parser(subp, raw):
    "Add a parser to the main subparser. "
    tmpp = subp.add_parser('show', help='display account status',
                           formatter_class=raw,
                           description=textwrap.dedent(DESC))
    tmpp.add_argument('type', type=str, choices=['users', 'keys'],
                      help='information type')


def show_keys():
    """Display your public key and your git key. """
    display = sys.stdout.write
    host = socket.gethostname()
    master = os.environ['USER']
    alias = git.config('host.alias')
    id_key, git_key = ssh.get_keys()
    id_key = ssh.get_public_key(id_key)
    display('# ID_RSA:\n')
    display('%s %s@%s - %s\n' % (id_key, master, host, alias))
    git_key = ssh.get_public_key(git_key)
    display('# GIT_KEY:\n')
    display('%s %s@%s - %s - git\n' % (git_key, master, host, alias))


def show_users():
    """Display all the users that have access to your account. """
    users, pending, unknown = ssh.read_authorized_keys()
    disp = sys.stdout.write
    disp('\n')
    emails = sorted(users.keys())
    for user in emails:
        disp('  %s:\n' % user)
        for key, content in users[user].items():
            disp('    ...%s: %s, %s, %s\n' % (key[-6:], content[1],
                                              content[0], content[2]))
        disp('\n')
    keys = pending.keys()
    if keys:
        disp('There are [%d] pending requests:\n\n' % len(keys))
        data = [(key[-6:], pending[key][0]) for key in keys]
        data = sorted(data, key=lambda x: x[1])
        for item in data:
            disp('  ...%s: %s\n' % item)
        disp('\n')
    if unknown:
        msg = 'There are [%d] unknown entries in ~/.ssh/authorized_keys:\n\n'
        disp(msg % len(unknown))
        for item in unknown:
            match = RE_LINE.match(item)
            if match:
                disp('  ...%s: %s\n' % (match.group('key')[-6:],
                                        match.group('desc')))
            else:
                disp('  NO MATCH ON: ...%s\n' % item[-10:])
        disp('\n')


def run(arg):
    """Run command. """
    func = {
        'keys': show_keys,
        'users': show_users,
    }
    func[arg.type]()



""" Promus core

This package contains the object that executes commands on your
behave as well as other utility functions.

"""

from promus.core.ssh import (
    make_key,
    get_keys,
    get_public_key,
    read_config,
    write_config,
    read_authorized_keys,
    write_authorized_keys,
)
from promus.core.git import (
    config,
    describe,
    repo_name,
    local_path,
    remote_path,
    make_hook,
    init,
    parse_dir,
    parse_acl,
    check_acl,
    read_acl,
    parse_profile,
    check_profile,
    read_profile,
    file_in_path,
    has_access,
    file_match,
    clone,
)
from promus.core.util import (
    is_exe,
    external_executables,
    check_promus_dependencies,
    user_input,
    encrypt_to_file,
    decrypt_from_file,
    wrap_msg,
    make_dir,
    parse_list,
    tokenizer,
    merge_lines,
    strip,
    send_mail,
)

import socket

# pylint: disable=R0902
class Promus(object):
    "An instance of this object manages the commands issued over ssh. "

    def __init__(self, script):
        # Host information
        self.host = socket.gethostname()
        self.alias = git.gconfig('host.alias')
        self.home = environ['HOME']
        self.master = environ['USER']
        self.master_name = git.gconfig('user.name')

        # Guest information
        self.guest = None
        self.guest_name = None
        self.cmd = None
        self.cmd_token = None
        self.cmd_name = None

        # Setting up log file
        self.path = '%s/.promus' % self.home
        util.make_dir(self.path)
        self.log_file = open('%s/promus.log' % self.path, 'a')
        script = realpath(script).replace(self.home, '~')
        self.log("INIT>> Running %s in %s" % (script, self.host))

        # Setting up functions based on command name
        self._exec = dict()
        self._exec['git-receive-pack'] = exec_git
        self._exec['git-upload-pack'] = exec_git

    def log(self, msg):
        "Write a message to the log file. "
        sys.stderr.write("[PROMUS]: %s\n" % msg)
        msg = '[%s:~ %s]$ %s\n' % (util.date(True), self.guest, msg)
        self.log_file.write(msg)

    def dismiss(self, msg, status):
        """Print msg to the standard error stream (sys.stderr), as
        well as to the log file and exit with the provided status. """
        self.log(msg)
        self.log_file.close()
        exit(status)

    def execute(self, cmd_name):
        """Execute one of the commands promus is allowed to execute.
        These are defined in the init function. If a command is not
        there then the deny function will be executed instead. """
        self._exec.get(cmd_name, deny)(self)

    def exec_cmd(self, cmd, verbose=False):
        "Run a subprocess and return its output and errors. "
        self.log("EXEC>> %s" % cmd)
        if verbose:
            out = sys.stdout
            err = sys.stderr
        else:
            out = PIPE
            err = PIPE
        process = Popen(cmd, shell=True, stdout=out, stderr=err)
        out, err = process.communicate()
        if verbose:
            with open('%s/promus.last' % self.path, 'w') as tmpf:
                tmpf.write("%s\n" % self.guest)
                tmpf.write("%s\n" % self.guest_name)
                tmpf.write("%s" % self.cmd)
        return out, err, process.returncode

    def _get_cmd(self):
        "Check to see if a command was given. Exit if it is not present. "
        if 'SSH_ORIGINAL_COMMAND' not in environ:
            msg = "GET_CMD-ERROR>> SSH_ORIGINAL_COMMAND not found."
            self.dismiss(msg, 1)
        self.cmd = environ['SSH_ORIGINAL_COMMAND']
        pattern = re.compile('.*?[;&|]')
        if pattern.search(self.cmd):
            msg = "GET_CMD-ERROR>> More than one command: %s" % self.cmd
            self.dismiss(msg, 1)
        self.cmd_token = self.cmd.split()
        self.cmd_name = self.cmd_token[0]

    def attend_last(self):
        """Reads the file containing the last guest and sets the
        guest info in order to proceed writing logs with that name.
        """
        with open('%s/promus.last' % self.path, 'r') as tmpf:
            info = tmpf.read()
        self.guest, self.guest_name, self.cmd = info.split('\n')
        self.cmd_token = self.cmd.split()
        self.cmd_name = self.cmd_token[0]

    def greet(self, guest, guest_name):
        "Handle the guest request. "
        self.log("GREET>> Connected as %s" % guest)
        self.guest = guest
        self.guest_name = guest_name
        self._get_cmd()
        if self.guest == self.master:
            self.exec_cmd(self.cmd, True)
        else:
            self.execute(self.cmd_name)
        self.dismiss("GREET>> done ...", 0)


